"""
    :created: 13.06.2018 by Jens Diemer
    :copyleft: 2018-2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import os
import tempfile
import unittest
from pathlib import Path

from django.conf import settings
from django.test import override_settings
from selenium.common.exceptions import NoSuchElementException

# https://github.com/jedie/django-tools
from django_tools.selenium.chromedriver import SeleniumChromiumTestCase, chromium_available
from django_tools.selenium.django import (
    SeleniumChromiumStaticLiveServerTestCase,
    SeleniumFirefoxStaticLiveServerTestCase,
)
from django_tools.selenium.geckodriver import SeleniumFirefoxTestCase, firefox_available
from django_tools.selenium.utils import find_executable
from django_tools.unittest_utils.assertments import assert_pformat_equal
from django_tools.unittest_utils.user import TestUserMixin


@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class ExampleChromiumTests(SeleniumChromiumStaticLiveServerTestCase):
    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")

        # FIXME: Depends on version?!?
        languages = self.driver.execute_script('return window.navigator.languages')
        language = languages[0]
        assert language in ('en', 'en-US')

        self.assert_equal_page_title("Log in | Django site admin")
        self.assert_in_page_source('<form action="/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()


@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class ExampleFirefoxTests(SeleniumFirefoxStaticLiveServerTestCase):
    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")

        self.assert_equal_page_title("Log in | Django site admin")
        self.assert_in_page_source('<form action="/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()


class SeleniumHelperTests(unittest.TestCase):
    def test_find_executable(self):
        with tempfile.NamedTemporaryFile(prefix="test_not_executable_", delete=False) as f:
            filepath = Path(f.name).resolve()
            self.assertTrue(filepath.is_file())
            name = filepath.name
            path = filepath.parent

            # File is not in PATH:
            with self.assertRaises(FileNotFoundError) as context_manager:
                find_executable(name)

            assert_pformat_equal(context_manager.exception.args, (f"Can't find '{name}' in PATH or None!",))

            old_path = os.environ["PATH"]
            try:
                # File is in PATH, but not executable:
                os.environ["PATH"] += f"{os.pathsep}{path}"
                with self.assertRaises(FileNotFoundError) as context_manager:
                    find_executable(name)

                assert_pformat_equal(
                    context_manager.exception.args, (f"{filepath} exists, but it's not executable!",)
                )

                # File is in PATH and executable:
                filepath.chmod(0x777)
                result = find_executable(name)
                assert_pformat_equal(result, filepath)
            finally:
                os.environ["PATH"] = old_path

            # Executable file is not in PATH, but can be found via extra_search_paths:
            result = find_executable(name, extra_search_paths=(path,))
            assert_pformat_equal(result, filepath)


class SeleniumTestsMixin:
    def test_login(self):

        self.assertTrue(settings.DEBUG)

        staff_data = self.get_userdata("staff")

        login_button_xpath = '//input[@value="Log in"]'

        url = self.live_server_url + "/admin/login/?next=/admin/"
        self.driver.get(url)

        self.assert_in_page_source("Log in | Django site admin")
        self.assert_not_in_page_source("errornote")
        self.assert_not_in_page_source("Please enter the correct username and password")

        self.assert_visible_by_id("id_username", timeout=2)
        self.assert_clickable_by_id("id_password", timeout=2)
        self.assert_clickable_by_xpath(login_button_xpath, timeout=2)

        try:
            username_input = self.driver.find_element_by_name("username")
            username_input.send_keys(staff_data["username"])

            password_input = self.driver.find_element_by_name("password")
            password_input.send_keys(staff_data["password"])

            login_button = self.driver.find_element_by_xpath(login_button_xpath)
            login_button.click()
        except NoSuchElementException as ex:
            self.fail(ex.msg)

        self.assert_no_javascript_alert()
        self.assert_not_in_page_source("Please enter the correct username and password")
        self.assert_not_in_page_source("errornote")
        self.assert_equal_page_title("Site administration | Django site admin")
        self.assert_in_page_source("<strong>staff_test_user</strong>")

        self.assert_visible_by_id("user-tools", timeout=2)
        self.assert_clickable_by_id("site-name", timeout=2)

    def test_admin_static_files(self):
        self.driver.get(self.live_server_url + "/admin/login/?next=/admin/")
        self.assert_in_page_source('href="/static/admin/css/base.css"')

        self.driver.get(self.live_server_url + "/static/admin/css/base.css")
        self.assert_in_page_source("margin: 0;")
        self.assert_in_page_source("padding: 0;")

    def test_local_storage_access(self):
        self.driver.get(self.live_server_url + "/")

        assert_pformat_equal(len(self.local_storage), 0)

        with self.assertRaises(KeyError) as cm:
            self.assert_local_storage_key_value(key="foo", value="bar")
        assert_pformat_equal(cm.exception.args[0], "foo")

        self.local_storage["bar"] = "foo"
        assert_pformat_equal(len(self.local_storage), 1)
        self.assert_local_storage_key_value(key="bar", value="foo")

        self.assertIn("bar", self.local_storage)

        assert_pformat_equal(repr(self.local_storage), "{'bar': 'foo'}")

        # There's no type conversion!
        self.local_storage["one"] = 1
        self.local_storage["t"] = True
        self.local_storage["f"] = False

        assert_pformat_equal(self.local_storage.items(), {"bar": "foo", "f": "false", "t": "true", "one": "1"})
        self.local_storage.clear()

        assert_pformat_equal(self.local_storage.items(), {})

        self.local_storage["foo"] = "bar"
        assert_pformat_equal(self.local_storage.items(), {"foo": "bar"})

        self.tearDown()  # tearDown should clear the local_storage
        assert_pformat_equal(self.local_storage.items(), {})


@override_settings(DEBUG=True)
@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class SeleniumChromiumAdminTests(TestUserMixin, SeleniumChromiumStaticLiveServerTestCase, SeleniumTestsMixin):
    def test_console(self):
        self.driver.get(self.live_server_url + "/")

        self.driver.execute_script("console.log('test console output 1');")
        self.assert_in_browser_log("test console output 1")

        self.driver.execute_script("console.log('test console output 2');")
        self.assert_in_browser_log("test console output 2")


@override_settings(DEBUG=True)
@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class SeleniumFirefoxAdminTests(TestUserMixin, SeleniumFirefoxStaticLiveServerTestCase, SeleniumTestsMixin):
    pass


class NonStaticLiveServerMixin:
    def test_google(self):
        self.driver.get("https://www.google.com/")
        self.assert_equal_page_title("Google")
        self.assert_in_page_source("Google Inc.")


@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class TestSeleniumChromiumNonStaticLiveServer(SeleniumChromiumTestCase, NonStaticLiveServerMixin):
    pass


@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class TestSeleniumFirefoxNonStaticLiveServer(SeleniumFirefoxTestCase, NonStaticLiveServerMixin):
    pass



#!/usr/bin/env python3

import os
import sys
from pathlib import Path

from django.core.management import BaseCommand, call_command

from django_tools_test_project.django_tools_test_app import migrations as test_app_migrations


print("sys.real_prefix:", getattr(sys, "real_prefix", "-"))
print("sys.prefix:", sys.prefix)


class Command(BaseCommand):
    """
    Expand django.contrib.staticfiles runserver
    """
    help = "Setup test project and run django developer server"

    def verbose_call(self, command, *args, **kwargs):
        self.stderr.write("_" * 79)
        self.stdout.write(f"Call {command!r} with: {args!r} {kwargs!r}")
        call_command(command, *args, **kwargs)

    def handle(self, *args, **options):

        if "RUN_MAIN" not in os.environ:
            # RUN_MAIN added by auto reloader, see: django/utils/autoreload.py

            migration_path = Path(test_app_migrations.__file__).parent
            for file_path in migration_path.glob('*.py'):
                if file_path.name != '__init__.py':
                    self.stdout.write(f'Remove migration file: {file_path}')
                    file_path.unlink()

            self.verbose_call("makemigrations")  # helpfull for developming and add/change models ;)
            self.verbose_call("migrate")

            # django.contrib.staticfiles.management.commands.collectstatic.Command
            self.verbose_call("collectstatic", interactive=False, link=True)

        self.verbose_call("runserver", use_threading=False, use_reloader=True, verbosity=2)



"""
    :created: 2015 by Jens Diemer
    :copyleft: 2015-2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import logging
import pprint

from selenium import webdriver
from selenium.webdriver import FirefoxOptions
from selenium.webdriver.firefox.service import Service
from selenium.webdriver.firefox.webdriver import DEFAULT_EXECUTABLE_PATH

from django_tools.selenium.base import LocalStorage, SeleniumBaseTestCase, assert_browser_language
from django_tools.selenium.utils import find_executable


log = logging.getLogger(__name__)


class SeleniumFirefoxTestCase(SeleniumBaseTestCase):
    """
    TestCase with Selenium and the Firefox WebDriver
    Note:
        * Needs 'geckodriver' executable! See README.creole for more info
        * It's without django StaticLiveServerTestCase

    usage e.g.:

        from django_tools.selenium.geckodriver import SeleniumFirefoxTestCase, firefox_available

        @unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
        class FirefoxTests(SeleniumFirefoxTestCase):

            def test_admin_login_page(self):
                self.driver.get(self.live_server_url + "/admin/login/")
                self.assert_equal_page_title("Log in | Django site admin")
                self.assert_in_page_source('<form action="/admin/login/" method="post" id="login-form">')
                self.assert_no_javascript_alert()

    see also: django_tools_tests/test_unittest_selenium.py
    """

    filename = DEFAULT_EXECUTABLE_PATH

    # Overwrite this in sub class, if needed:
    extra_search_paths = ()

    options = (
        "-headless",
    )
    desired_capabilities = {
        "loggingPrefs": {
            "browser": "ALL",
            "client": "ALL",
            "driver": "ALL",
            "performance": "ALL",
            "server": "ALL"
        }
    }

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        try:
            executable = find_executable(cls.filename, cls.extra_search_paths)
        except FileNotFoundError as err:
            log.exception('"%r" not found: %s', cls.filename, err)
        else:
            options = FirefoxOptions()

            for argument in cls.options:
                options.add_argument(argument)

            for key, value in cls.desired_capabilities.items():
                options.set_capability(key, value)

            options.set_preference('intl.accept_languages', 'en-US, en')

            log.debug('Browser options:\n%s', pprint.pformat(options.to_capabilities()))
            service = Service(
                executable_path=str(executable),
                log_path=f'{cls.filename}.log'
            )
            cls.driver = webdriver.Firefox(
                options=options,
                service=service,
            )

            # Test may fail, if a other language is activated.
            # So check this after startup:
            assert_browser_language(driver=cls.driver, languages=('en', 'en-US'))

            cls.local_storage = LocalStorage(cls.driver)


def firefox_available(filename=None):
    """
    :return: True/False if 'firefox-chromedriver' executable can be found

    >>> firefox_available("doesn't exists")
    False
    """
    if filename is None:
        filename = SeleniumFirefoxTestCase.filename

    try:
        executable = find_executable(filename, SeleniumFirefoxTestCase.extra_search_paths)
    except FileNotFoundError as err:
        log.error("Firefox is not available: %s", err)
        return False

    log.debug(f"Firefox found here: {executable}")
    return True



"""
    Test django_tools.unittest_utils.django_command

    :copyleft: 2017-2021 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from pathlib import Path

from bx_py_utils.test_utils.snapshot import assert_text_snapshot
from django.core.cache import cache
from django.core.management import call_command
from django.test import TestCase

import django_tools
import django_tools_test_project
from django_tools.unittest_utils.assertments import assert_pformat_equal
from django_tools.unittest_utils.django_command import DjangoCommandMixin
from django_tools.unittest_utils.stdout_redirect import StdoutStderrBuffer
from django_tools.unittest_utils.user import TestUserMixin


REPO_PATH = str(Path(django_tools.__file__).parent.parent.parent)
MANAGE_DIR = Path(django_tools_test_project.__file__).parent


class CommandTestCase(TestUserMixin, DjangoCommandMixin, TestCase):
    def test_help(self):
        output = self.call_manage_py(["--help"], manage_dir=MANAGE_DIR)

        self.assertIn("[django]", output)
        self.assertIn("[django_tools]", output)
        self.assertIn("nice_diffsettings", output)

        self.assertNotIn("Traceback", output)
        self.assertNotIn("ERROR", output)

        assert 'Use settings: ' in output
        assert REPO_PATH in output
        output = output.replace(REPO_PATH, '...')
        assert_text_snapshot(got=output)

    def test_list_models(self):
        output = self.call_manage_py(
            ["list_models"], manage_dir=MANAGE_DIR,  # debug=True
        )
        assert "existing models in app_label.ModelName format:" in output
        output = output.split("existing models in app_label.ModelName format:", 1)[1]

        assert '01 - admin.LogEntry' in output
        assert 'INSTALLED_APPS....:' in output
        assert 'Apps with models..:' in output

        assert REPO_PATH not in output
        assert_text_snapshot(got=output)

    def test_nice_diffsettings(self):
        output = self.call_manage_py(["nice_diffsettings"], manage_dir=MANAGE_DIR)

        self.assertIn("\n\nSETTINGS_MODULE = 'django_tools_test_project.settings.test'\n\n", output)
        self.assertIn("\n\nINSTALLED_APPS = ('django.contrib.auth',\n", output)

        self.assertNotIn("Traceback ", output)  # Space after Traceback is important ;)
        self.assertNotIn("ERROR", output)

        assert 'Use settings: ' in output
        assert REPO_PATH in output
        output = output.replace(REPO_PATH, '...')
        assert_text_snapshot(got=output)

    def test_environment(self):
        usernames = ",".join(self.UserModel.objects.values_list("username", flat=True).order_by("username"))
        assert_pformat_equal(usernames, "normal_test_user,staff_test_user,superuser")

    def test_no_username_given(self):
        with StdoutStderrBuffer() as buff:
            call_command("permission_info")
        output = buff.get_output()

        self.assertIn("All existing users are:", output)
        self.assertIn("normal_test_user, staff_test_user, superuser", output)
        self.assertIn("(3 users)", output)

        self.assertNotIn("Traceback", output)
        self.assertNotIn("ERROR", output)

        assert REPO_PATH not in output
        assert_text_snapshot(got=output)

    def test_wrong_username_given(self):
        with StdoutStderrBuffer() as buff:
            call_command("permission_info", "not existing username")
        output = buff.get_output()

        self.assertIn("Username 'not existing username' doesn't exists", output)

        self.assertNotIn("Traceback", output)
        self.assertNotIn("ERROR", output)

        assert REPO_PATH not in output
        assert_text_snapshot(got=output)

    def test_normal_test_user(self):
        with StdoutStderrBuffer() as buff:
            call_command("permission_info", "normal_test_user")
        output = buff.get_output()

        self.assertIn("Display effective user permissions", output)
        self.assertIn("is_active    : yes", output)
        self.assertIn("is_staff     : no", output)
        self.assertIn("is_superuser : no", output)

        self.assertIn("[ ] auth.add_user", output)
        self.assertIn("[ ] sites.delete_site", output)

        self.assertNotIn("[*]", output)  # normal user hasn't any permissions ;)
        self.assertNotIn("missing:", output)  # no missing permissions

        self.assertNotIn("Traceback", output)
        self.assertNotIn("ERROR", output)

        assert REPO_PATH not in output
        assert_text_snapshot(got=output)

    def test_update_permissions(self):
        with StdoutStderrBuffer() as buff:
            call_command("update_permissions")
        output = buff.get_output()

        self.assertIn("Create permissions for:", output)
        self.assertIn(" * auth", output)
        self.assertIn(" * django_tools_test_app", output)

        self.assertNotIn("Traceback", output)
        self.assertNotIn("ERROR", output)

        assert REPO_PATH not in output
        assert_text_snapshot(got=output)

    def test_clear_cache(self):

        cache.set("key_foo", "bar", 30)
        assert_pformat_equal(cache.get("key_foo"), "bar")

        with StdoutStderrBuffer() as buff:
            call_command("clear_cache")
        output = buff.get_output()

        assert_pformat_equal(cache.get("key_foo"), None)

        self.assertIn("Clear caches:", output)
        self.assertIn("Clear 'LocMemCache'", output)
        self.assertIn("done.", output)

        self.assertNotIn("Traceback", output)
        self.assertNotIn("ERROR", output)

        assert REPO_PATH not in output
        assert_text_snapshot(got=output)
